package server

import (
	"errors"
	"fmt"
	lru "github.com/hashicorp/golang-lru"
	odysseus "github.com/odysseus/odysseus-protocol/gen/proto/go/base/v1"
	omanager "github.com/odysseus/odysseus-protocol/gen/proto/go/nodemanager/v1"
	"github.com/redis/go-redis/v9"
	log "github.com/sirupsen/logrus"
	"sync"
	"time"
)

var (
	ErrWorkerExist = errors.New("worker exist")
)

type dispatchTask struct {
	task  *odysseus.TaskContent
	errCh chan error
}

type Worker struct {
	quit       chan interface{}
	taskCh     chan *dispatchTask
	resultCh   chan *omanager.SubmitTaskResult
	uuid       int64
	publicKey  string
	recentTask *lru.Cache
	stream     omanager.NodeManagerService_RegisterWorkerServer
}

type WorkerManager struct {
	rdb       *redis.Client
	heartBeat map[int64]int64
	hbRwLock  sync.RWMutex

	workers  map[int64]*Worker
	wkRwLock sync.RWMutex
	quit     chan struct{}
}

func NewWorkerManager(rdb *redis.Client) *WorkerManager {
	return &WorkerManager{
		heartBeat: make(map[int64]int64),
		workers:   make(map[int64]*Worker),
		quit:      make(chan struct{}),
		rdb:       rdb,
	}
}

func (wm *WorkerManager) UpdateHeartBeat(uuid int64) {
	wm.hbRwLock.Lock()
	defer wm.hbRwLock.Unlock()
	wm.heartBeat[uuid] = time.Now().Unix()
}

func (wm *WorkerManager) UpdateStatus(worker *Worker) {

}

func (wm *WorkerManager) GetHeartBeat(uuid int64) int64 {
	wm.hbRwLock.RLock()
	defer wm.hbRwLock.RUnlock()
	return wm.heartBeat[uuid]
}

func (wm *WorkerManager) GetWorker(uuid int64) *Worker {
	wm.wkRwLock.RLock()
	defer wm.wkRwLock.RUnlock()

	return wm.workers[uuid]
}

func (wm *WorkerManager) AddNewWorker(uuid int64, worker omanager.NodeManagerService_RegisterWorkerServer) (*Worker, error) {
	wm.wkRwLock.Lock()
	defer wm.wkRwLock.Unlock()

	if _, exist := wm.workers[uuid]; exist {
		return nil, ErrWorkerExist
	}
	w := &Worker{
		taskCh:   make(chan *dispatchTask),
		resultCh: make(chan *omanager.SubmitTaskResult),
		uuid:     uuid,
		stream:   worker,
	}
	taskCache, err := lru.New(100)
	if err != nil {
		return nil, err
	}
	w.recentTask = taskCache
	wm.workers[uuid] = w
	go wm.handleWorkerMsg(w)

	return w, nil
}

type Callback func(err error) bool

func (wm *WorkerManager) manageWorker(worker *Worker) error {
	heartBeatDuration := time.Second * 10
	workerCheckDuration := heartBeatDuration * 3

	heartBeatTicker := time.NewTicker(heartBeatDuration)
	defer heartBeatTicker.Stop()

	workerCheckTicker := time.NewTicker(workerCheckDuration)
	defer workerCheckTicker.Stop()

	statusTicker := time.NewTicker(time.Second * 10)
	defer statusTicker.Stop()

	deviceInfoTicker := time.NewTicker(time.Second * 10)
	defer deviceInfoTicker.Stop()

	deviceUsageTicker := time.NewTicker(time.Second * 10)
	defer deviceUsageTicker.Stop()

	for {
		var msg = new(omanager.ManagerMessage)
		var callback = Callback(func(err error) bool {
			// do nothing
			return true
		})

		select {
		case <-wm.quit:
			gb := new(omanager.ManagerMessage_GoodbyeMessage)
			gb.GoodbyeMessage = &omanager.GoodbyeMessage{}
			msg.Message = gb
		case <-workerCheckTicker.C:
			if time.Now().Unix()-wm.GetHeartBeat(worker.uuid) > int64(workerCheckDuration.Seconds()) {
				// remove worker
				close(worker.quit)
			}

		case <-heartBeatTicker.C:
			hb := new(omanager.ManagerMessage_HeartbeatRequest)
			hb.HeartbeatRequest = &omanager.HeartbeatRequest{
				Timestamp: uint64(time.Now().Unix()),
			}
			msg.Message = hb

		case <-deviceInfoTicker.C:
			deviceInfo := new(omanager.ManagerMessage_DeviceRequest)
			deviceInfo.DeviceRequest = &omanager.DeviceInfoRequest{}
			msg.Message = deviceInfo
			callback = func(err error) bool {
				if err == nil {
					deviceInfoTicker.Reset(time.Second * 180)
				}
				return true
			}

		case <-deviceUsageTicker.C:
			deviceUsage := new(omanager.ManagerMessage_DeviceUsage)
			deviceUsage.DeviceUsage = &omanager.DeviceUsageRequest{}
			msg.Message = deviceUsage
			callback = func(err error) bool {
				if err == nil {
					deviceUsageTicker.Reset(time.Second * 180)
				}
				return true
			}

		case <-statusTicker.C:
			status := new(omanager.ManagerMessage_StatusRequest)
			status.StatusRequest = &omanager.StatusRequest{}
			msg.Message = status
			callback = func(err error) bool {
				if err == nil {
					statusTicker.Reset(time.Second * 120)
				}
				return true
			}

		case dtask, ok := <-worker.taskCh:
			if !ok {
				return nil
			}
			task := dtask.task
			taskMsg := new(omanager.ManagerMessage_PushTaskMessage)
			taskMsg.PushTaskMessage = &omanager.PushTaskMessage{
				TaskId:    task.TaskId,
				TaskType:  task.TaskType,
				Workload:  0,
				TaskCmd:   task.TaskCmd,
				TaskParam: task.TaskParam,
			}
			msg.Message = taskMsg
			callback = func(err error) bool {
				if err == nil {
					worker.recentTask.Add(task.TaskId, task)
				}

				select {
				case dtask.errCh <- err:
				default:
					// err ch is invalid
				}
				return true
			}
		}

		if msg != nil {
			err := worker.stream.Send(msg)
			if err != nil {
				log.WithError(err).Error("send message to worker failed")
			}
			callback(err)
		}
	}

	return nil
}
func (wm *WorkerManager) handleWorkerMsg(worker *Worker) {
	for {
		select {
		case <-wm.quit:
			return
		case <-worker.quit:
			return
		default:
			wmsg, err := worker.stream.Recv()
			if err != nil {
				log.WithError(err).WithField("worker", worker.uuid).Error("recv msg failed")
				close(worker.quit)
				return
			}
			switch msg := wmsg.Message.(type) {
			case *omanager.WorkerMessage_GoodbyeMessage:
				worker.quit <- msg.GoodbyeMessage.Reason
				close(worker.taskCh)
				return
			case *omanager.WorkerMessage_SubmitTaskResult:
			case *omanager.WorkerMessage_HeartbeatResponse:
			case *omanager.WorkerMessage_Status:
			case *omanager.WorkerMessage_DeviceInfo:
			case *omanager.WorkerMessage_DeviceUsage:
			default:
				log.WithField("worker", worker.uuid).Error(fmt.Sprintf("unsupport msg type %T", msg))
			}
		}
	}
}
